"""           
@author: estevez schwarz/lamour

    Copyright (C) 2020

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from scipy.optimize import minimize
from scipy.linalg import pinv,norm
from numpy import dot,zeros, size, array,sqrt,append,array_str, set_printoptions,eye,NaN
from InitDAE.DAEanalysis.NumericalException import MinimizationNotSuccessful, LittleNumberOfTaylorCoefficients,\
    SingularDAE
from numpy.linalg.linalg import matrix_rank
from InitDAE.toolboxAD.MyUTPM import MyUTPM
from InitDAE.toolboxLA.svdinfo import OrthogonalProjectorAlongIm,OrthogonalProjectorOntoKer
from InitDAE.DAEanalysis.DerivativeArray import TaylorDAE
        
def func(X,P,alpha,ex):
    """ Objective function  ||P(x_0-alpha)||_2, where X = (x_0,x_1,...)
        :param X - ndarray of Taylor variable
        :param P - orthogonal projector
        :param alpha - 'initial value'
        :param ex
    """
    n=size(P,0)
    x0=X[:n]
    y=dot(P,x0-alpha)
    yty=dot(y,y)
    logging.getLogger("consi").log(20,'\t\t\t\t func: sqrt(yty) {0:10.2e}'.format(sqrt(yty)))
    if ex.info_output[1]>'1':
        print ('\tfunc: sqrt(yty) {0:10.2e}'.format(sqrt(yty)))#,y    
    return sqrt(yty)

def func_deriv(X,P,alpha,ex):
    """ Derivative of objective function """
    n=size(P,0)
    #print 'dimension ',n
    deriv=zeros(X.size)
    x0=X[:n]
    #print 'x0 ',x0
    y=dot(P,x0-alpha)
    yty=dot(y,y)
    #print 'func_deriv: yty',yty,y
    if abs(sqrt(yty))<2.3e-16:
        deriv[:n]=zeros(n)
    else:
        deriv[:n]=y/sqrt(yty)
    #print 'func_deriv: deriv ',deriv
    return deriv

#from AD_examples.Symbolic.test_twoPendula import JacobianDerivativeArray_sym as JacobianDerivativeArray, derivativeArray_sym as derivativeArray
#from AD_examples.Symbolic.symbolic_test import JacobianDerivativeArray_sym as JacobianDerivativeArray, derivativeArray_sym as derivativeArray
#from AD_examples.Symbolic.symbolic_test import JacobianDerivativeArray_sym, derivativeArray_sym
#from AD_examples.Symbolic.test_twoPendula import JacobianDerivativeArray_sym, derivativeArray_sym

def derivativeArray(X,taylorObject,ti,K,dim):
    """
    derivativeArray creates the derivativeArray as ndarray 
    """
    T0Taylor = MyUTPM.taylor_init(K, ti)
    #print 'in derivativeArray T0Taylor ',T0Taylor
    XTaylor=MyUTPM(zeros((K, 1, dim)))
    for i in range(K):
        XTaylor.data[i,0]=X[i*dim:(i+1)*dim]
        
    #print 'in derivativeArray XTaylor',XTaylor
    f_Taylor=taylorObject.f_taylor(XTaylor,T0Taylor)
    fx=f_Taylor.flatten()
    Km1mdim=(K-1)*dim
    # check whether u exists
    if hasattr(taylorObject.example,'function_u') and taylorObject.example.function_u:
        ux=taylorObject.example.u(X[:dim])
        length_u=ux.shape[0]
        if length_u > dim:
            print ('\nderivativeArray: ----- number of additional requirements (example.u) = ',length_u,' greater then dimension of the DAE = ',dim)
            exit()
        else:
            fx[Km1mdim:Km1mdim+length_u]=ux
            Km1mdim=Km1mdim+length_u
            # the length of f increases by length_u
    norm_fx= norm(fx[:Km1mdim])
        
    if taylorObject.info_output[1]>'1':
        print ('derivativeArray: norm(f_Taylor) {0:10.2e}'.format(norm_fx),' ',end="")#,'\n\n',f_Taylor.flatten()
    logging.getLogger("consi").log(20,'\nderivativeArray: norm(f_Taylor) {0:10.2e}'.format(norm(fx[:Km1mdim])))
    taylorObject.normDerivativeArray = norm_fx
    ###################### symbolic test #################################
    #DerArray=derivativeArray_sym(X,taylorObject,ti,K,dim)
    #print ' K= ',K
    #print  "\nderivative array fx \n",fx[:Km1mdim]
    #print  '\nderivative array DerArray \n',DerArray
    #exit()
    ######################################################################  
    return fx[:Km1mdim]
    
def JacobianDerivativeArray(X,taylorObject,ti,K,dim):  
    '''
    JacobianDerivativeArray creates the Jacobian matrix of the derivativeArray as two dimensional ndarray 
    '''
    T0Taylor = MyUTPM.taylor_init(K, ti)
    XTaylor=MyUTPM(zeros((K, 1, dim)))
    for i in range(K):
        XTaylor.data[i,0]=X[i*dim:(i+1)*dim]
        
    Coeff_Matrix,dim,f_Taylor=taylorObject.coefficent_matrix_optimization(T0Taylor,XTaylor)
        
    if hasattr(taylorObject.example,'function_u') and taylorObject.example.function_u:
       
        u_x=taylorObject.Jacobian_matrix_u(X[:dim])
        Coeff_Matrix=append(Coeff_Matrix,append(u_x,zeros((u_x.shape[0],dim*(K-1))),axis=1),axis=0)
        rju=matrix_rank(Coeff_Matrix[:,:])
        dim_rju=size(Coeff_Matrix,0)
        #print 'JacobianDerivativeArray: rank, dimension  ',rju,dim_rju
        if rju!=dim_rju:
            print ('\n\nJacobianDerivativeArray: ----- Additional condition u maybe not consistent! No full row rank.')
    ###################### symbolic test #################################
    #Jacobi=JacobianDerivativeArray_sym(X,taylorObject,ti,K,dim)
    #print '\nJacobimatrizen norm(CoeffMatrix-Jacobi) ',norm(Coeff_Matrix-Jacobi)
    #print '\nJacobimatrizen CoeffMatrix \n',Coeff_Matrix
    #print '\nJacobimatrizen Jacobi \n',Jacobi
    #exit()
    ######################################################################  
    return Coeff_Matrix

def constraint(taylorObject,ti,K,dim):
    """
    constraint delivers the dictionary containing the derivative array of order K of the example (in taylorObject)
    at timepoint ti.
    """
    cons = ({'type': 'eq',
              'fun' : derivativeArray,
              'jac' : JacobianDerivativeArray,
              'args':(taylorObject,ti,K,dim)})
    
    return cons  

def consistentValues(example,TaylorOrder,alpha,x0,ti,augmentation=False,ftol=1e-8,disp=True,maxiter=100,info_output='01'):
    """
    consistentValues computes consistent values of the DAE example in the timepoint ti such that || P(x(ti)-alpha) ||_2 -> min! using the modul minimize with the method SLSQP
        :param: 
            * **example** - object of example class
            * **TaylorOrder** - integer, order of the highest derivative of Taylor coefficients
            * **alpha** - reference vector as ndarray of size dim = dimension of the DAE
            * **x0** - ndarray containing initial guess of length dim*TaylorOrder
            * **ti** - timepoint
            * **augmentation** 
                * if False: using the TaylorOrder
                * if True:  augmentation of number of Taylor coefficients starting with 1 up to TaylorOrder+1
            * **ftol** = 1e-8 - (parameter for python module *minimize*)
            * **disp** = True - (parameter for python module *minimize*)
            * **maxiter** = 100 - (parameter for python module *minimize*): maximal number of iterations
            * **info_level** = 0 - integer, higher level implies more information, 0 ... 3
                                   If info_level = 0 no output with the exception of error messages.\n
                                   If info_level > 0 then two log-files are generated: *example_name*\ _array_\ *date*\ .log and *example_name*\ _consi_\ *date*\ .log \n
                                   *array* contains Jacobian matrices, projectors, ranks, SVD values etc.\n
                                   *consi* contains values related to the computation of consistent initial values and the solution in full accuracy (16 digits)
        :raised:
            * **exceptions**: 
                * MinimizationNotSuccessful
                * ValueError
        :return: 
            * **solution** as ndarray of length dim*(TaylorOrder+1) containing values for x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(TaylorOrder)`\ (t0)/(TaylorOrder)!
                **Caution**: Only  x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(s)`\ (t0)/s!  are consistent for s = TaylorOrder+1-index (see [1])
            * **TaylorEx** Object of class TaylorDAE of the example (to be used in index_Condition) 
    """
    info_level=max(info_output[0],info_output[1])
    K = TaylorOrder+1
    dimension=alpha.shape
    dim=dimension[0]
    # check of the right length of x0
    if x0.size != TaylorOrder*dim:
        print ('consistentValues: Wrong length of x0, should be TaylorOrder*alpha.size = ',TaylorOrder*dim,' but it is',x0.size,'.\n')
        raise ValueError('Wrong length of x0')
    if augmentation:
        X=x0[:dim]
        for K_temp in range(2,K+1):
            K_tempmdim=K_temp*dim
            if K_temp==K:
                X=append(X,zeros(dim))
            else:
                X=append(X,x0[K_tempmdim-dim:K_tempmdim])
            #X[:dim]=alpha
            if info_level>'2':
                print ('K_temp= ',K_temp,'\n\nX = \n',X.reshape(K_temp,dim))
            if hasattr(example,'svdtol'):
                tolexample=example.svdtol
            else:
                tolexample=None    
            TaylorEx = TaylorDAE(example, K_temp,tol=tolexample,info_output=info_output)
            if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                if info_level>'0':
                    print ('consistentValues: A function u with additional conditions is used.')
                logging.getLogger("consi").log(20,'\nconsistentValues: A function u with additional conditions is used.')
                
            CoeffMatrix=JacobianDerivativeArray(X,TaylorEx,ti,K_temp,dim)
            G0=CoeffMatrix[:dim,dim:2*dim]
            Gm=pinv(G0)
            P0=dot(Gm,G0)
            
            logging.getLogger("array").log(20,'\nconsistentValues: P0 =\n\n'+str(P0))
            projPi=projectorPi(CoeffMatrix,P0,dim)
            if info_level > '2':
                print ('consistentValues: Pi= \n',projPi)
            
            res = minimize(func, X, args=(P0,alpha,TaylorEx), jac=func_deriv, constraints=constraint(TaylorEx,ti,K_temp,dim),\
                            method='SLSQP', options={'disp': disp,'maxiter':maxiter,'ftol':ftol})
            if res.success and TaylorEx.normDerivativeArray > 10*ftol:
                raise MinimizationNotSuccessful()
            if disp:
                print ('minimize success =',res.success,' status =',res.status)
            if res.success:
                X=res.x
            else:
                raise MinimizationNotSuccessful()
    else:
        X=append(x0,zeros(dim))
        if hasattr(example,'svdtol'):
            tolexample=example.svdtol
        else:
            tolexample=None
        TaylorEx = TaylorDAE(example, K,tol=tolexample,info_output=info_output)
        if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                if info_level>'0':
                    print ('consistentValues: A function u with additional conditions is used.')
                logging.getLogger("consi").log(30,'\nconsistentValues: A function u with additional conditions is used.')    
                
        CoeffMatrix=JacobianDerivativeArray(X,TaylorEx,ti,K,dim)
        G0=CoeffMatrix[:dim,dim:2*dim]
        Gm=pinv(G0)
        P0=dot(Gm,G0)
        
        logging.getLogger("consi").log(10,'\nConsistentValues: info_output[1] = '+info_output[1])
        logging.getLogger("array").log(20,'\nConsistentValues: P0 =\n\n'+str(P0))
       
        projPi=projectorPi(CoeffMatrix,P0,dim)
        if info_level > '2':
            print ('consistentValues: Pi= \n',projPi)
        
        res = minimize(func, X, args=(P0,alpha,TaylorEx), jac=func_deriv, constraints=constraint(TaylorEx,ti,K,dim),\
                       method='SLSQP', options={'disp': disp,'maxiter':maxiter,'ftol':ftol})
        
        if res.success and TaylorEx.normDerivativeArray > 10*ftol:
            print ('consistentValues: Norm of the derivative array to big = {:9.6e}'.format(TaylorEx.normDerivativeArray))
            raise MinimizationNotSuccessful()
        if disp:
            print ('minimize success =',res.success,' status =',res.status)
        if not res.success:
            raise MinimizationNotSuccessful()
    logging.getLogger("consi").log(30,'\nConsistentValues: '+res.message)
    resString=''
    for step in range(K):
        set_printoptions(precision=16)
        resString=resString+array_str(res.x[step*dim:step*dim+dim], max_line_width=dim*28)+'\n'
    equals=    '\n===============================================================\n'
    logging.getLogger("consi").log(30,'\n'+equals+'ConsistentValues: Computed consistent values at t = '+str(ti)+equals+'\n'+resString )
    return res.x,TaylorEx

def index_Condition(TaylorEx,dimOfDAE,TaylorVector,ti):
    """
    index_Condition calculates the index of the DAE and the matrix condition number of the "Jacobian" of the derivative array
        :param: 
            * **TaylorEx** - object of class TaylorDAE of the example
            * **dimOfDAE** - dimension of the DAE
            * **TaylorVector** - TaylorVector as ndarray (e.g. solution of consistentValues)
            * **ti** - timepoint
        :return: 
            * **index** - index of the DAE
            * **cond** - condition number
            * **s** - level of s-fullness
    """
    
    K=int(TaylorVector.size/dimOfDAE)
    #print ('type of K ',type(K))
    T0Taylor = MyUTPM.taylor_init(K, ti)
    #print 'in derivativeArray T0Taylor ',T0Taylor
    XTaylor=MyUTPM(zeros((K, 1, dimOfDAE)))
    for i in range(K):
        imdimOfDAE=i*dimOfDAE
        XTaylor.data[i,0,0:dimOfDAE]=TaylorVector[imdimOfDAE:imdimOfDAE+dimOfDAE]
    try:
        index,cond,s=TaylorEx.IndexAndCondition(T0Taylor, XTaylor)
        
    except SingularDAE as ex:
        print (ex)
        print ('index_Condition: Singular DAE, index and condition number could not be computed.')
        raise
    except LittleNumberOfTaylorCoefficients as ex:
        #print ex
        print ('index_Condition: Index and condition number could not be computed.')
        raise
    return index,cond,s

def projectorPi(G,P,n):
    """
    projectorPi computes the projector Pi - see paper
        G - coefficent matrix of constraints
        P - projector P
        n - dimension of DAE
    """
    #print 'in projectorPi'
    W1,rS,S1,Sr,Sr1,Send=OrthogonalProjectorAlongIm(G[:,n:])
    aux=dot(W1,G[:,:n])
    Q=eye(n)-P
    aux1=dot(aux,Q)
    W2,rS,S1,Sr,Sr1,Send=OrthogonalProjectorAlongIm(aux1)
    aux1=dot(W2,aux)
    aux=append(Q, aux1, axis=0)
    Pi,rS,S1,Sr,Sr1,Send,used_tol=OrthogonalProjectorOntoKer(aux)
    #print 'out projectorPi'
    return Pi
    
def consistentValuesIndex(example,TaylorOrder,alpha,x0,t0,augmentation=False,ftol=1e-8,disp=False,maxiter=100,info_output='01'):
    """
    consistentValuesIndex wraps the moduls consistentValues and index_Condition.
         :param: 
            * **example** - object of example class
            * **TaylorOrder** - integer, order of the highest derivative of Taylor coefficients
            * **alpha** - reference vector as ndarray of size dim = dimension of the DAE
            * **x0** - ndarray containing initial guess of length dim*TaylorOrder
            * **t0** - timepoint
            * **augmentation** = False
                * if False: using the TaylorOrder
                * if True:  augmentation of number of Taylor coefficients starting with 1 up to TaylorOrder+1
            * **ftol** = 1e-8 - (parameter for python module *minimize*)
            * **disp** = False - (parameter for python module *minimize*)
                         If *disp* = True also results of the iteration of the minimizer are displayed.
            * **maxiter** = 100 - (parameter for python module *minimize*): maximal number of iterations
            * **info_output** - string 'ab'
            * **info_output** =  higher level implies more information, '0' ... '3'
                                   If info_output = '0' no output with the exception of error messages.\n
                                   If 'a' > '0' then log-file *example_name*\ _array_\ *date*\ .log is generated.\n
                                       *array* contains Jacobian matrices, projectors, ranks, SVD values etc.\n
                                   If 'b' > '0' then log-file *example_name*\ _consi_\ *date*\ .log \n
                                       *consi* contains values related to the computation of consistent initial values and the solution in full accuracy (16 digits)
         :return: 
            * **solution** as ndarray of length dim*(TaylorOrder+1) containing values for x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(TaylorOrder)`\ (t0)/(TaylorOrder)!
                **Caution**: Only  x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(s)`\ (t0)/s!  are consistent for s = TaylorOrder+1-index (see [1])
            * **index** of the DAE in the computed consistent point
            * **cond** - condition number of the index computation
            * **computation_robust** - If True -> The relation TaylorOrder+1-index-s==0 is fulfilled
    """
    info_level=info_output[1]
    dimension=alpha.shape
    n=dimension[0]
    stars='*'*57+'\n'
    try:
        if info_level>'1':
            print ('\n',stars,'consistentValuesIndex: Computation of consistent solution\n',stars)
        solution,TaylorEx=consistentValues(example,TaylorOrder,alpha,x0,t0,augmentation=augmentation,ftol=ftol,disp=disp,maxiter=maxiter,\
                                                                                                        info_output=info_output)
        if info_level>'0':
            print ('\n\n',stars,'consistentValuesIndex: Consistent solution at t = ',t0,'\n',stars,'\n',solution.reshape(TaylorOrder+1,n))
        index,cond,s=index_Condition(TaylorEx,n,solution,t0)
        if s==-1:
            logging.getLogger("consi").log(30,'\nconsistentValuesIndex: DAE seems to be singular!\n'\
                                         +'                       The computed minimization solution does not represent consistent values,\n'\
                                         +'                       index and condition number could not be computed.')
            return solution,None,cond,False
        if index==0:
            s+=1 # see DAE_TaylorDAE.Check_s_full
        
        computation_robust= TaylorOrder+1-index-s==0   
        if info_level>'0':    
            print ('\nconsistentValuesIndex: index = ',index,' D = ',(TaylorOrder+1),'  s = ',s, ' (D - index  == s ?) = ',computation_robust,\
                  '\n                       condition number of the derivative array considered in the final step of index determination = {0:10.2e}'.format(cond))                             
        logging.getLogger("consi").log(20,'\nconsistentValuesIndex: \n'+'Number of Taylor coefficents = '\
                                     +str(TaylorOrder+1)+'  index mu = '+str(index)+'  cond = {0:10.2e}'.format(cond)\
                                     +'  Coefficient matrix is '+str(s)+'-full')
        if TaylorOrder+1-index != s:
            if info_level>'0':
                messagepart1='consistentValuesIndex: The relation D-mu == s is not fulfilled. Check the condition number and the validity of the index determination!!!!!\n'
                messagepart2=' or in the array-file and maybe adapt the parameter svdtol in the example.'
                if info_output[0]<'2':
                    message=messagepart1+\
                            '                       Set info_output[0] > "1" and check the singular value S(r) in the terminal output'+messagepart2
                            
                else:
                    message=messagepart1+\
                            '                       Check the singular value S(r) in the terminal output'+messagepart2
                print (message)
                logging.getLogger("consi").log(30, message)       
        return solution,index,cond,computation_robust
    
    except MinimizationNotSuccessful as ex:
        print (ex,'\n\nconsistentValuesIndex: InitDAE cannot compute a consistent solution! MinimizationNotSuccessful')
        if 'TaylorEx' in locals(): # this is the case when consistentValues is aborted by Exception
            if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                print ('consistentValuesIndex: ----- Additional function u is used, it maybe that the condition(s) in function u are not performable.')
        return None,NaN,NaN,None
    except ValueError as ex:
        print (ex,'\n\nconsistentValuesIndex: InitDAE cannot compute a consistent solution! ValueError')
        return None,NaN,NaN,None
    except LittleNumberOfTaylorCoefficients as ex:
            print ('consistentValuesIndex: ',ex)
            logging.getLogger("consi").log(30,'\nconsistentValuesIndex: Number of Taylor coefficients too small!\n'\
                                         +'                       The computed minimization solution does not represent consistent values,\n'\
                                         +'                       index and condition number could not be computed.')
            raise
            #return None,NaN,NaN,None
        
if __name__=="__main__":    

    from AD_examples.KroneckerTestDoku.example_Kronecker_proper import Kronecker
    ex = Kronecker()
    
    n = 5
    K = 10
    
    t0=0.
    alpha=array([1.,1.,1.,1.,1.])
    
    print ('Solution by minimization')
 
    print ('alpha = ',alpha)
    ftola=1e-14
    x0=zeros(K*n)
    x0[:n]=alpha
    try:
        solution,index,cond,robust=consistentValuesIndex(ex,K,alpha,x0,t0,ftol=ftola,info_output='01',disp=True)
    except LittleNumberOfTaylorCoefficients as ex:
        print (ex)
  
    
