'''
Created on 13.01.2016

@author: lamour

    Copyright (C) 2020

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
#import algopy # for the method argument
from algopy import UTPM, zeros, Function, CGraph
from numpy import concatenate,shape
class MyUTPM(UTPM):
         
    def __str__(self):
        return UTPM.__str__(self)

    def diff(self):
        """ 
        differentiation for UTPM objects by shifting the coefficients
        """
        derivative=self.shift(-1) # 
        D = self.data.shape[0]
        
        for i in range(1,D-1):
            
            derivative.data[i,...]=(i+1)*derivative.data[i,...]
            
        return derivative   
         
    @staticmethod   
    def taylor_init(D,t0): 
        """ 
        initialization of independent variable t at t0
        """
        t=UTPM(zeros((D,1)))
        t.data[0,0,]=t0
        if D>=2:
            t.data[1,0,]=1
        return t
        
    def tracing_A_D_B(self,d,f,t):
        '''
        The function traces the functions d and f to compute the Jacobians D, A and B
        '''
        def F(z,t,length_y,length_x): 
            """ Auxiliary function to compute the Jacobian matrices"""
            
            y = z[:length_y]          # Splitting into the components y and x
            x = z[length_y:]
            out = f(y,x,t)
            return out 
          
        cg = CGraph() # start tracing
        fx = Function(self)
        tt = Function(t)
        fy = d(fx,tt)
        cg.trace_off()   # stop tracing
        cg.independentFunctionList = [fx]
        cg.dependentFunctionList = [fy]   
        
        fyy=d(self,t)
        fy_s=fyy.diff() # derivative of  d as UTPM-object

        cgAB = CGraph()
        yx=concatenate((fy_s.data,self.data),axis=2) # Assemble of y and x to one object
        yyx=UTPM(yx)
        fyyx=Function(yyx)

        length_y=fy_s.data.shape[2]
        length_x=self.data.shape[2]
        
        Fyxtt=F(fyyx,tt,length_y,length_x)
        cgAB.trace_off()
        cgAB.independentFunctionList = [fyyx] # 
        cgAB.dependentFunctionList = [Fyxtt] 
        
        return cg,cgAB,Fyxtt.x.data
        
    def tracing_A_B(self,f,t):
        '''
        The function traces the function f to compute the Jacobians A and B
            :param:
                * **f** - function f
                * **t** - time point
            :return: 
                * **cgAB** - computational graph of f
                * **Fyxtt.x.data** - ndarray of Taylorcoefficients of f
        '''
        def F(z,t,length_y,length_x): 
            """ Auxiliary function to compute the Jacobian matrices"""
            
            y = z[:length_y]   # Splitting into the components y and x
            x = z[length_y:]
            out = f(y,x,t)
            return out 
        cgAB = CGraph() # start tracing
        tt = Function(t)
        y_s=self.diff() # derivative of  d as UTPM-object
        yx=concatenate((y_s.data,self.data),axis=2) # Assemble of y and x to one object
        yyx=UTPM(yx)
        fyyx=Function(yyx)
        
        length_y=y_s.data.shape[2]
        length_x=self.data.shape[2]
        Fyxtt=F(fyyx,tt,length_y,length_x)
        cgAB.trace_off() # stop tracing
        cgAB.independentFunctionList = [fyyx] # 
        cgAB.dependentFunctionList = [Fyxtt] 
        return cgAB,Fyxtt.x.data 
    
    def A_B(self, cgAB, f, t):
        """ 
        Calculation of the Jacobians A and B
            :param:
                * **cgAB** - computational graph
                * **f** - function f
                * **t** - UTPM object - taylor_init of t
            :return:
                * **A** - Taylor object of Jacobian df/dx'
                * **B** - Taylor object of Jacobian df/dx
        """
        y_s=self.diff() # derivative of  d as UTPM-object
        yx=concatenate((y_s.data,self.data),axis=2) # Assemble of y and x to one object
        yyx=UTPM(yx)
        length_y=y_s.data.shape[2]
        #print 'A_B: cgAB ',cgAB
        AB= cgAB.jacobian(yyx)
        A=AB[:,:length_y]
        B=AB[:,length_y:]
        return (A,B)
    
    def A_D_B_new(self, cg, cgAB, d, f, t):
        """ 
        Calculation of the Jacobians A, D and B
            :param:
                * **cg** - computational graph of f
                * **cgAB** - computational graph of A, B
                * **d** - function d of proper formulation
                * **f** - function f
                * **t** - UTPM object - taylor_init of t
            :return: 
                * **D** - Taylor object of Jacobian matrix dd/dx
                * **A** - Taylor object of Jacobian matrix df(z,x,t)/dz
                * **B** - Taylor object of Jacobian matrix df(z,x,t)/dx
        """
        
        D= cg.jacobian(self)

        # computation of the Jacobian matrices A = df/dy and B = df/dx

        fyy=d(self,t)
        fy_s=fyy.diff() # derivative of  d as UTPM-object
        yx=concatenate((fy_s.data,self.data),axis=2) # Assemble of y and x to one object
        yyx=UTPM(yx)
        length_y=fy_s.data.shape[2]
        AB= cgAB.jacobian(yyx) 

        A=AB[:,:length_y]
        B=AB[:,length_y:]
        return (D,A,B)
    
if __name__ == "__main__":
    
    from AD_examples.MonitoringOfSingularity.example_ind2_singAt_t_0 import Example_ind2_singAt_t_0
    ex = Example_ind2_singAt_t_0()
    
    n=3
    K=3
    
    ti = 4.
    #Taylor element of time ti
    Time = MyUTPM.taylor_init(K, ti)
    # zt is Taylor in t_i 
    zt = MyUTPM(zeros((K, 1, n)))
    # set of values
    zt.data[0, 0, 0] = 1
    zt.data[0, 0, 1] = 2
    zt.data[0, 0, 2] = 3
    #set of values of first Taylor coefficient
    zt.data[1, 0, 0] = -1
    zt.data[1, 0, 1] = -2
    zt.data[1, 0, 2] = -3
    
    print ('How does MyUTPM compute the Jacobian matrices? Here demonstrated for the case type_of_DAE="proper". ')
    print ("t=\n", Time,"\n\nzt=\n", zt)
    cg,cgAB,Fyxttxdata=zt.tracing_A_D_B(ex.d,ex.f,Time)
    (D, A, B) = zt.A_D_B_new(cg,cgAB,ex.d, ex.f, Time)

    print ("A=\n", A, "\n D=\n", D, "\nB=\n", B)

    print ("================= different time = 1 and zt - check the entries in matrix B and compare. ===========================")
    Time.data[0,0]=1.
    zt.data[0, 0, 0] = -2.3
    zt.data[0, 0, 1] = 45.3
    #z = zt[0, :]
    (D, A, B) = zt.A_D_B_new(cg, cgAB, ex.d, ex.f, Time)
    
    print ("Time = \n",Time,"\n\nzt=",zt)
    print ("A=\n", A, "\n\n D=\n", D, "\n\nB=\n", B)
   