# -*- svdinfo: utf-8 -*-
"""
Created on Thu Mar 05 09:07:30 2015

@author: estevez schwarz

    Copyright (C) 2020

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy.linalg import svd,norm
from numpy import finfo,dot,shape,zeros

def CompCondSing(M,tol=None):
    '''
    CompCondSing computes the ratio between the largest and the smallest nonzero singular value of a matrix M and returns some related information.
        :param: 
            * **M** - Matrix M
            * **tol** - tolerance for the rank decision (optional). The computation of the default tolerance corresponds to linalg.matrix_rank
        :return: 
            * **cond** - S[0]/S[rankM-1]: ratio between the largest and the smallest nonzero singular value of M
            * **rankM** - considered rank  of the matrix M
            * **S[0]** - largest singular value
            * **S[rankM-1]** - smallest singular value considered as positive
            * **S[min(len(S)-1,rankM)]** - largest singular value considered as zero
            * **S[-1]** - smallest singular value
            * **tol** - used tolerance
    '''  
    if M.ndim > 2:
        raise TypeError('CompCondSing: array should have 2 or fewer dimensions')
    if M.ndim < 2:
        return int(not all(M==0))
    S = svd(M, compute_uv=False)
    if tol is None:
        used_tol = norm(M,'fro') * max(M.shape) * finfo(S.dtype).eps
    else:
        used_tol = norm(M,'fro') * max(M.shape) *tol
    rankM= sum(S > used_tol)
    # modification
    return (S[0]/S[rankM-1],rankM,S[0],S[rankM-1],S[min(len(S)-1,rankM)],S[-1],used_tol) 
  
  
def RankSigma(S,tol=None):
    '''
    RankSigma computes the rank of the diagonal matrix Sigma (S) from the SVD
        :param:  
            * **S** - Matrix S
            * **tol** - tolerance for the rank decision (optional). The computation of the default tolerance corresponds to linalg.matrix_rank
        :return: 
            * **rank** - rank(S)
            * **tol** - used tolerance
    '''  
    # Code from linalg.matrix_rank
    if tol is None:
        used_tol = norm(S) * max(S.shape) * finfo(S.dtype).eps
    else:
        used_tol = norm(S) * max(S.shape) * tol
    return sum(S > used_tol),used_tol

    
def OrthogonalProjectorAlongIm(A,tol=None):
    '''
    OrthogonalProjectorAlongIm  computes the orthogonal projector along im(A) for a given matrix A
        :param:  
            * **A** - Matrix A
        :return: 
            * **W** - resulting orthogonal projector
            * **rS** - considered rank of the matrix A (and corresponding S from SVD)
            * **S1** - largest singular value of A
            * **Sr** - smallest singular value considered as positive
            * **Sr1** - largest singular value considered as zero
            * **Send** - smallest singular value        
    '''  
    dim=shape(A)      
    #print 'in OrthogonalProjectorAlongIm'  
    UTr,rS,S1,Sr,Sr1,Send,used_tol= OrthogonalBasisProjectorAlongIm(A,tol=tol)
    if rS== dim[0]:
        W=zeros((dim[0],dim[0]))
    else:
        W=dot(UTr.T,UTr)
    return [W,rS,S1,Sr,Sr1,Send]
    
    
def OrthogonalBasisProjectorAlongIm(A,tol=None):
    '''
    OrthogonalBasisProjectorAlongIm computes the orthogonal basis to build the projector along im(A) for a given matrix A
        :param:  
            * **A** - Matrix A
        :return: 
            * **Ur.T** - resulting orthogonal basis
            * **rS** - considered rank of the matrix A (and corresponding S from SVD)
            * **S[0]** - largest singular value of A
            * **S[rS-1]** - smallest singular value considered as positive
            * **S[min(len(S)-1,rS)]** - largest singular value considered as zero
            * **S[-1]** - smallest singular value         
    '''
    #print 'in OrthogonalBasisProjectorAlongIm'
    dim=A.shape
    (U,S,VH)=svd(A) 
    rS,used_tol=RankSigma(S,tol=tol)
    if rS == dim[0]:
        Ur=zeros((dim[0],1))
    else:
        Ur=U[:,rS:]
    return  Ur.T,rS,S[0],S[rS-1],S[min(len(S)-1,rS)],S[-1],used_tol


def OrthogonalProjectorOntoKer(A,tol=None):
    '''
    OrthogonalProjectorOntoKer  computes the orthogonal projector onto ker(A) for a given matrix A
        :param: 
            * **A** - Matrix A
        :return: 
            * **Q** - resulting orthogonal projector
            * **rS** - considered rank of the matrix A (and corresponding S from SVD)
            * **S1** - largest singular value of A
            * **Sr** - smallest singular value cosidered as positive
            * **Sr1** - largest singular value considered as zero
            * **Send** - smallest singular value           
    '''     
    dim=shape(A)
   
    VTr,rS,S1,Sr,Sr1,Send,used_tol=OrthogonalBasisProjectorOntoKer(A,tol=tol)
    if rS == dim[1]:
        Q=zeros((dim[1],dim[1]))
    else:
        Q=dot(VTr.T,VTr)
    return [Q,rS,S1,Sr,Sr1,Send,used_tol]


def OrthogonalBasisProjectorOntoKer(A,tol=None):
    '''
    OrthogonalBasisProjectorOntoKer computes the orthogonal basis to build the projector onto ker(A) for a given matrix A
        :param: 
            * **A** - Matrix A
        :return: 
            * **VTr** - resulting orthogonal basis
            * **rS** - considered rank of the matrix A (and corresponding S from SVD)
            * **S[0]** - largest singular value of A
            * **S[rS-1]** - smallest singular value cosidered as positive
            * **S[min(len(S)-1,rS)]** - largest singular value considered as zero
            * **S[-1]** - smallest singular value             
    '''   
    dim=shape(A)
    #print 'OrthogonalBasisProjectorOntoKer:  dim A = ',dim
    (U,S,VT)=svd(A) 
    rS,used_tol=RankSigma(S,tol=tol)
    if rS == dim[1]:
        VTr=zeros((dim[1],dim[1])) 
    else:
        VTr=VT[rS:,:]
    #print 'OrthogonalBasisProjectorOntoKer:  tol = ',used_tol,' len(S)= ',len(S),' rS = ',rS,' \nS= ',S    
    return (VTr,rS,S[0],S[rS-1],S[min(len(S)-1,rS)],S[-1],used_tol)
