'''
Created at 20.05.2014/ last change 25.03.2020

@author: estevez/lamour

    Copyright (C) 2020

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
import logging
import numpy
from numpy import  shape, append, loadtxt, array,resize,ones, reshape, sign, zeros as npzeros,isnan,NaN,dot, size,sqrt,array_str, set_printoptions
from numpy.linalg import norm
from scipy.special import factorial
from algopy import zeros
from InitDAE.toolboxAD.MyUTPM import MyUTPM
from InitDAE.DAEinitialization.consistentValues import consistentValuesIndex,index_Condition,JacobianDerivativeArray,constraint
from InitDAE.DAEanalysis.DerivativeArray import TaylorDAE
from scipy.optimize import minimize
from InitDAE.DAEanalysis.NumericalException import MinimizationNotSuccessful,LittleNumberOfTaylorCoefficients,SingularDAE,MinimalStepsize
from scipy.linalg import pinv
########################
    
    
class TaylorIntegration(TaylorDAE):
    '''
    From the TaylorDAE class, the TaylorIntegration object inherits the attributes and  has the additionally attributes
        :param:
            * **info_output** - additional component 'c', string 'abc' controls the information output, 0 no output up to 3 most information
                a - level for linear algebra, like SVD, rank determination etc.\n
                b - level for calculation of consistent initial values\n
                c - level for integration\n
                standard - '001'
            * **method** - integration method
                = 0: Explicit Taylor K-index=ke, ki=0\n
                = 1: Fully implicit Taylor K-index=ki, ke=0\n
                = 2: Two-halfstep-method K-index=ke=ki\n
                = 3: HOP method for ki=ke\n
                = 4: HOP method for ki-1=ken\n
                = 5: HOP method for ki-2=ke     
            * **FOR INTERNAL USE ONLY** 
            * **Kc** - initial value = None
            * **ki** - initial value = None
            * **ke** - initial value = None
            * **h** - initial value = None
            * **Xlast** - initial value = None
            * **HOPweightsE** - initial value =None
            * **HOPweightsI** - initial value =None
            * **index** - initial value =NaN       
    '''
    def __init__(self,example,K,tol=None,info_output='001',method=0):
    
        TaylorDAE.__init__(self,example,K,tol=tol,info_output=info_output)
        #self.__init__(example,K,tol=tol,info_output=info_output)
        self.method = method
        if not 0 <= method <= 5:
            print ('TaylorIntegration_init: No such METHOD ( = ',method,'). method = 0,...,5 only')
            exit()
      
        self.Kc = None
        self.ki = None
        self.ke = None
        self.h = None
        self.Xlast= None
        self.HOPweightsE=None
        self.HOPweightsI=None
        self.index=NaN
        
    def weights_HOP(self):
        method=self.method
        # method == 3: HOP method for ki=ke
        # method == 4: HOP method for ki=ke+1
        # method == 5: HOP method for ki=ke+2
        if method>2:
            # HOP-methods
            # ki=self.K-int(self.index)# otherwise inconsistent values are used
            ki=self.K-int(self.index)
            if method == 3:
                ke=ki
            elif method == 4:
                # L-stable
                if ki>0:
                    ke=ki-1
                else:
                    print ('weights_HOP: Increase K by 1 to use method=4')
                    exit()
            elif method == 5:
                # L-stable
                if ki>1   :
                    ke=ki-2
                else:
                    print ('weights_HOP: Increase K by 2 to use method=5')
                    exit()
            
        #print 'HOPweightE',self.HOPweightsE
        #print 'HOPweightI',self.HOPweightsI
        
        self.HOPweightsE=npzeros((ke+1))
        self.HOPweightsE[0]=1.
        for j in range(ke):
            self.HOPweightsE[j+1]=self.HOPweightsE[j]*(ke-j)/(ke+ki-j)
            
        self.HOPweightsI=npzeros(ki+1)
        self.HOPweightsI[0]=1.   
        for j in range(ki):
            self.HOPweightsI[j+1]=self.HOPweightsI[j]*(ki-j)/(ke+ki-j)
            
        self.ki=ki
        self.ke=ke
        
            
    def check_solutionpath(self,x,a,b,time_points=100):
        '''
        check_solutionpath computes the condition of the Taylor coefficient matrix in every point of the  solution x in the interval [a,b]
            :param:
                * **x** - function of t, which delivers Taylorcoefficents too. (X=algopy.zeros(dimension,dtype=t) )
                * **a,b** - time interval of interest
                * **time_points** - number of sampling points
            :return:
                * **condition** - ndarray(time_points+1,2), condition[:,0] = list of timepoints, condition[:,1] = list of condition numbers
                * **index** - ndarray(time_points+1,1) = list of DAE index
        '''
        # condition is a vector containing the condition numbers, the stepsizes at every time point and the condition of the pencil
        
        h=(b-a)/float(time_points)
        condition=npzeros((time_points+1,2))
        index_list=npzeros(time_points+1)
        ti=a-h
        for i in range(time_points+1):
            ti+=h
            condition[i,0]=ti
            tti=MyUTPM.taylor_init(self.K,ti)
            
            xti=x(tti)
            dim=xti.shape[0]
            y=reshape(xti.data, (dim*self.K))
            try:
                index,cond,s=index_Condition(self,dim,y,ti)
            except SingularDAE as ex:
                print ('check_solutionpath: ',ex)
                cond=1e15
                index=-1
            condition[i,1]=cond
            index_list[i]=index

        return condition,index_list

    def check_solution_along_path(self, path,ftol=1e-10):
        ''' 
        check_solution_along_path delivers the index and the condition number of a DAE along a by a list given path
            :param:
                * **path** - String containing the path of a file with list of solutions 
                    structure of every row of the ASCII-file: timepoint x_1 x_2 ... x_n
            :return:     
                * **t_list** - list of timepoints from file
                * **index** -  list of DAE index at timepoints
                * **condition** - list of condition numbers at timepoints 
        '''
        K=self.K
        ex=self.example
        info_level=self.info_output[2]
      
        liste_zzt = adapt_solution_list(path,K)
        
        liste=  loadtxt(path)
        dim=shape(liste)
        #print 'check_solution_along_path: dim liste ', dim
        # number of timepoints
        time=dim[0]
        n=dim[1]-1
       
        condM=npzeros(time)
        index=npzeros(time)
        X0=npzeros(K*n)
        
        for i in range(time): # main loop 
             
            ti=float(liste[i,0])
            alpha=liste[i,1:]
            #print 'check_solution_along_path: ti, alpha ',ti, alpha
            function_u_changed=False
            if hasattr(ex,'function_u') and ex.function_u:
                        ex.function_u=False
                        function_u_changed=True
            X0[:n]=alpha
            solution,index[i],condM[i],robust=consistentValuesIndex(ex,K,alpha,X0,ti,ftol=ftol,disp=False,info_output=self.info_output[:2])
            if solution is not None:
                X0=solution[:-n]
            else:
                if info_level>'0':
                    print ('check_solution_along_path: No convergence at {:8.3e} - solution is "None"'.format(ti))
                if i!=0:
                    if isnan(index[i]):
                        #index[i]=index[i-1]
                        index[i]=-1
                    if isnan(condM[i]):
                        condM[i]=condM[i-1]
                else:
                    if isnan(index[i]):
                        index[i]=-1
                    if isnan(condM[i]):
                        condM[i]=0.0
        if function_u_changed:
            ex.function_u=True
        return liste[:,0],index,condM


    
    def Taylor_integration_timeList(self,ti_list,x0,ftol=1e-14,max_iterations=20):
        '''
        Taylor__integration_timeList integrates a DAE with given steps in the parameter ti_list
            :param:
                * **ti_list** - numpy.ndarray containing [t0,t1,...,ti,...,T]
                * **x0** - initial guess
                * **ftol** = 1e-14
                * **max_iterations** = 20
            :return:
                * **ti_list** - list of successful t points (mostly identical with input list)
                * **xout** - solution at timepoints in t-_list
                * **index** - array of DAE index at every timepoint
                * **condM** - array of condition number at every timepoint
        ''' 
        NumSteps=ti_list.shape[0] # NumSteps = Number of points
        n=len(x0)
        
        xout=npzeros((NumSteps,n))   
        #----------------------------------------------------------        
        output='Taylor_integration_timeList: max_iterations= {:4d}, ftol= {:9.3e}, Integration method= {:2d}'\
                .format(max_iterations,ftol,self.method)+'\n'
        logging.getLogger("integr").log(30,output)

        index=npzeros(NumSteps)
        condM=npzeros(NumSteps)
###########   
        self.h=ti_list[1]-ti_list[0]
        dim=shape(x0)
        initial_x0=npzeros(self.K*dim[0])
        initial_x0[:dim[0]]=x0
        #print 'Taylor_integration_constant_h t=', t
        
        for k in range(NumSteps):
           if self.info_output[2]>'0': 
               print ('Taylor_integration_timeList: h= {:6.3e}  t= {:6.3e}'.format(self.h,ti_list[k]))
          
           if k==0:  
               solution,index[k],condM[k],robust=consistentValuesIndex(self.example,self.K,x0,initial_x0,ti_list[k],\
                                                                       augmentation=True,ftol=ftol,disp=False,maxiter=max_iterations,info_output=self.info_output)
               # if index[k] is NaN:
               if isnan(index[k]):
                   print ('Taylor_integration_timeList: index=NaN')
                   exit()
               if solution is None:
                   print ('Taylor_integration_timeList: No consistent initial values at t=t0')
                   exit()
               self.index=index[k]
               if self.method>2:
                   self.weights_HOP()
               # the example uses an additional function u. After initialization it will be switched off.
               if hasattr(self.example,'function_u') and self.example.function_u:
                   self.example.function_u=False
           else:
               h=ti_list[k]-ti_list[k-1]
               self.h=h
               x0=self.Taylor_Step(solution,h)
               alpha=self.Taylor_Method_alpha(solution,n,h)
             
               solution,index[k],condM[k],robust=self.consistentStepIndex(self.K,alpha,x0,ti_list[k],augmentation=False,ftol=ftol,disp=False,\
                                                                          maxiter=max_iterations,info_output=self.info_output)
               if solution is None:
                   print ('-----Taylor_integration_timeList: No consistent values computed. Singularity?')
                   return ti_list[:k],xout,index,condM
               if isnan(index[k]):
                    print ('-----Taylor_integration_timeList: consistentStepIndex fails at ',ti_list[k])
                    index[k]=-1
                    self.index=index[k]
               else:
                    self.index=index[k]
            
           OK=robust
           if OK>=0:
               xout[k,:]=solution[:dim[0]].copy()
               
    ################ Provisorisch:
               if k==0:
                   DAE_index=index[k] 
               elif DAE_index !=index[k] :
                   print ('Taylor_integration_timeList: Index changed at t=', ti_list[k],'index_old = ',DAE_index,'  index_new = ',index[k]) 
                   DAE_index=index[k]
    ################            
           else:
               print ('Taylor_integration_timeList: Computation of consistent initial values is not robust.')
               xout[k,:]=solution[:dim[0]].copy() 
           self.logging_integration_solution(ti_list[k],xout[k,:])
        
        return ti_list,xout,index,condM
    
    def Taylor_integration_adaptiveStep(self,t0,T,alpha,max_iterations=20,tol=1e-6,hmin=1e-8,K_start=2,disp=False,RobustIntegration=True,ftol=1e-8,hstart=-1.):
        '''
        Taylor_integration with adaptive stepsize. The idea consists in the computation of consistent initial values,\n
        the estimation of a relevant stepsize, a step using the Taylor expansion and so on.
            :param:
                * **t0** - initial time point
                * **T** - end of the integration interval
                * **x0** - initial guess of x(t0), see also computation of consistent initial values
                * **max_iterations** - maximal number of iterations in minimizer 
                * **tol** - rel. integration accuracy
                * **hmin** - minimal stepsize
                * **ftol** - accuracy of consistent initial value computation
                * **hstart** - initial stepsize, if < 0 estimated stepsize is used.
            :return:
                * **xout** - ndarray with solution at time points
                * **tlist** - list of timepoints
                * **index** - list of DAE index at timepoints
                * **condM** - list of condition number at timepoints
        ''' 
#----------------------------------------------------------        
        output='Taylor_integration_adaptiveStep: max_iterations= {:4d}, tol= {:9.3e}, hmin= {:9.3e}, disp={:}, RobustIntegration= {:}, ftol= {:9.3e}, hstart= {:9.3e}, Integration method= {:2d}'\
                .format(max_iterations,tol,hmin,disp,RobustIntegration,ftol,hstart,self.method)+'\n'
        logging.getLogger("integr").log(30,output)
#---------------------------------------------------------- 
        eps=2**(-52)
        h=hmin*1e6 # for start the while loop only
        #h=hmin*1.01 # Test Diana
        hmax=abs(T-t0)/10
        #hmax=abs(T-t0)/10000 # Test Diana
        #print 'adaptive: Achtung hmax'
        stepsizereduction=False # for stepsize control
        reduction_steps=0 # counter for the reduction steps
        dim=shape(alpha)[0]
        x0=npzeros(dim*self.K)
        x0[:dim]=alpha
        t=t0
        example=self.example
        TaylorOrder=self.K
        info_level=self.info_output[2]
        if T==t0: # case only to compute consistent initial values
            direction=1.0
        else:
            direction=sign(T-t0)
        # using of hstart
        if hstart > 0.0:
            if hstart <= hmin:
                print ('adaptiveStepsize: hstart <= hmin, hmin changed to hstart/2')
                hmin=hstart/2
            h=hstart
        h=h*direction
        print ('adaptiveStepsize: initial stepsize ',h)
        while abs(h)>=hmin:
         
            if t==t0:
                try:
                    solution,index,condM,computation_robust=consistentValuesIndex(example,TaylorOrder,alpha,x0,t,augmentation=True,\
                                                                                  ftol=ftol,disp=disp,maxiter=max_iterations,info_output=self.info_output)
                    if solution is not None and not isnan(index):
                        if RobustIntegration and not computation_robust:
                            print ('adaptiveStepsize: The computation at this step was not robust.')
                        ok=1
                        print ('adaptiveStepsize: initialization at t=t0 index =',index)
                        self.index=index
                        print ('Taylor_integration_adaptiveStep: TAYLOR-METHOD',self.method)  
                        if self.method>2:
                            self.weights_HOP()
                        # the example uses an additional function u. After initialization it will be switched off.
                        if hasattr(example,'function_u') and example.function_u:
                            example.function_u=False
                    else:
                        ok=-3
                except MinimizationNotSuccessful as e: #IncreasingTaylorCoefficient,e: # das muss noch abgefangen werden
                    print (e)
                    ok=-3
                except ValueError as e:
                    print (e)
                    ok=-3
                except LittleNumberOfTaylorCoefficients as e:
                    #print e
                    ok=-3
            else:
                try:
                    #print 'initialization at t=',t
                    
                    #x0[:dim]=x0[:dim]#*1.0001 # the minimizer has problems in minimums -> therefore a small perturbation
                    solution,index,condM,computation_robust=self.consistentStepIndex(TaylorOrder,alpha,x0,t,augmentation=False,ftol=ftol,disp=disp,\
                                                                                     maxiter=max_iterations,info_output=self.info_output)
                    if solution is not None and not isnan(index):
                        ok=1 #normal successful case
                        if RobustIntegration and not computation_robust:
                            print ('adaptiveStepsize: The computation at this step was not robust.') 
                            ok=-2 # robust integration required
                    else:
                        ok=-2 # stepsize reduction
                except MinimizationNotSuccessful as e:
                    print (e)
                    ok=-2
                except ValueError as e:
                    print (e)
                    ok=-2
                except LittleNumberOfTaylorCoefficients as e:
                    #print e
                    ok=-3       
            # initialization of index_old for the case that ok==0 (no integration necessary)
            #                of the outputlist of xout and tlist
            if ok>=0:
                
                if t==t0:
                    xout=solution[:dim].copy()
                    tlist=numpy.array([t0])
                    indexList=numpy.array([index])
                    condList=numpy.array([condM])
                    condM_old=condM
                    index_old=index
                    DAE_index=index
                else:
                    xout=append(xout,solution[:dim])
                    tlist=append(tlist,t)
                    indexList=append(indexList,index)
                    if ok==1:
                        condList=append(condList,condM)
                    else: # case ok=0, no new condM computed
                        condList=append(condList,condM_old)
                    condM_old=condM
                    index_old=index    
            if ok >= 0:
                if DAE_index !=index :
                    print ('adaptiveStepsize: INDEX CHANGED at t= {:12.6e} from index_old = {:3d} to index = {:3d}'.format(t,DAE_index,index))
                    DAE_index=index
################     
                X0_old=solution.copy()
                self.Xlast=X0_old
                t_old=t
                if 0<=index<2:
                    estimation_coefficient=self.K-1
                else:   
                    estimation_coefficient=self.K-index
                if estimation_coefficient<=0:
                    print ('adaptiveStepsize: Number of Taylor coefficents ( = ',self.K,') too small for DAE_Index =',index)
                    break
                #
                # h last used stepsize, hh estimated new stepsize
                #
                norm_xk=norm(solution[(estimation_coefficient-1)*dim:estimation_coefficient*dim])
                #print 'adaptiveStepsize: solution \n',solution
                        
# new stepsize estimation
                if (norm_xk > 10*eps):
                    hh=(tol/norm_xk)**(1./estimation_coefficient) 
                    estimation=True
                    #print 'adaptiveStepsize: hh, norm_xk, estimation_coefficient',hh, norm_xk, estimation_coefficient

                    hh=min(hh,2*abs(h))*direction
                    
                    if (abs(hh) < hmin ):
                        print ('adaptiveStepsize:  Stepsize smaller than hmin = ',hmin,' at t = ',t,'.')
                        break
                        #ok=-5;    
                else:
                    hh=min(hmax, 2*abs(h))*direction
                        #text_int='enlargement (double) h= {:9.3}'.format(hh)#+' h= '+str(h)+' hmax= '+str(hmax)+' norm = '+str(norm_xk)
                    estimation=False
# case stepsize reduction                    
                if abs(hh) < abs(h):
                    # case that reduction behaves like halving the stepsize
                    if t==t0 and hstart > 0.0:
                        print ('adaptiveStepsize: first stepsize h = {:9.3e}'.format(hstart)+' is used')
                        h=hstart*direction
                    else:
                        if 2*abs(hh) <= abs(h):
                            reduction_steps +=1
                        h=abs(hh)*direction
                        text_int='reduction   to h= {:9.3e}'.format(h)
                else:
                    if t==t0: # setting to the estimated stepsize at the first step
                        if hstart < 0.0:
                            h=min(abs(hh),hmax)*direction
                            print ('adaptiveStepsize: estimated first stepsize h={:9.3e}'.format(h)+' is used')
                        else:
                            h=hstart*direction
                            print ('adaptiveStepsize: first stepsize h = {:9.3e}'.format(h)+' is used')
                        #h=5*hmin*direction # Test Diana
                        #print 'ACHTUNG: Reduction of h_start to 5*hmin'
                        text_int='initialization h= {:9.3e}'.format(h)
                    else:
# case stepsize enlargement - also after reduction                        
                        if reduction_steps > 0:
                            if stepsizereduction:
                                hh=abs(h) # old stepsize for one step
                                text_int='unchanged      h= {:9.3e}'.format(hh*direction)
                            else:
                                hh=min(abs(hh),2*abs(h))
                                text_int='doubling    to h= {:9.3e}'.format(hh*direction)
                            h=hh*direction
                            
                            reduction_steps -=1
                        else:
                            hnew=abs(hh)*direction
                            if abs(hh)!=abs(h):
                                text_int='enlargement to h= {:9.3e}'.format(hnew)
                            else:
                                text_int='unchanged      h= {:9.3e}'.format(hnew)
                            h=hnew    
                          
                stepsizereduction=False   
                #h=min(h,tf-t);
                if abs(T-t) < abs(h):
                    if abs(T-t)<2*eps: # the final point is reached
                        self.logging_integration_solution(t,solution[:dim])
                        text_int=' time = {:9.3e}'.format(t)
                        text_int='adaptiveStepsize:                                     '+text_int
                        print (text_int)
                        break
                    h=T-t
                    text_int='to reach T     h= {:9.3e}'.format(h)
                
                if(abs(h)>=hmax):
                    h=hmax*direction
                    text_int='maximal    hmax = {:9.3e}'.format(h)
                text_int+=' time = {:9.3e}'.format(t)
                text_int='adaptiveStepsize: stepsize '+text_int
                self.logging_integration_solution(t,solution[:dim])

                print (text_int)
                self.h=h
                alpha=self.Taylor_Method_alpha(solution,dim,h)
                x0=self.Taylor_Step(solution,h)
                t=t+h
                #T0.data[0,0]=t
                
            elif(abs(h)/2 > hmin and t!=t0):
                reduction_steps +=1
                h=h/2
                stepsizereduction=True
                print ('adaptiveStepsize: stepsize bisection to   h= {:9.3e} time = {:9.3e}'.format(h,t_old))
                self.h=h
                alpha=self.Taylor_Method_alpha(X0_old,dim,h)
                x0=self.Taylor_Step(X0_old,h)
                t=t_old+h
                #print ' t= ',t
            else:
                if t==t0:
                    xout=None  # initialization for mistake free output only
                    tlist=[]
                    indexList=[]
                    condList=[]# initialization for mistake free output only
                    print ('adaptiveStepsize:  No convergence at starting point, use better initial guess')
                else:
                    print ('adaptiveStepsize:  Stepsize smaller then hmin = ',hmin,' at t = ',t,'.')
                    if abs(T-t) > 0.0:
                        yxout=reshape(xout,(len(tlist),len(alpha)))
                        return yxout,tlist,indexList,condList
                    #raise MinimalStepsize()
                break
            
        if t!=t0:
            yxout=reshape(xout,(len(tlist),len(alpha)))
            return yxout,tlist,indexList,condList
        else:
            return None,tlist,indexList,condList
    
    def logging_integration_solution(self,t,x):
        """ logging_integration_solution writes a tuple of t,x(t) to a logging file """
        output='{:18.15e}'.format(t)+'  '
        for i in range(len(x)):
            output+='{:18.15e}'.format(x[i])+'  '
        #print 'logging_integration_solution: output',output
        logging.getLogger("integr").log(30,output)
                    
    def Taylor_integration_constant_h(self,t0,x0,T,NumSteps,ftol=1e-14,max_iterations=20):
        '''
        Taylor__integration_constant_h integrates a DAE over the interval [t0,T] with NumSteps ( = number of intervals )
            :param:
                * **t0** -
                * **x0** - initial value
                * **T** -
                * **NumSteps**
                * **ftol**
                * **max_iterations**
            :return:
                * **xout** - ndarray with solution at time points
                * **ti_list** - list of timepoints
                * **index** - list of DAE index at timepoints
                * **condM** - list of condition number at timepoints
        ''' 
        output='Taylor_integration_constant_h: t0= {:9.3e}, T= {:9.3e}, NumSteps={:5d}'\
                .format(t0,T,NumSteps)
        logging.getLogger("integr").log(30,output)
        h=(T-t0)/NumSteps
        
        ti_list=npzeros(NumSteps+1)
        ti_list[0]=t0
        for i in range(NumSteps-1):    
            ti_list[i+1]=ti_list[i]+h
        ti_list[NumSteps]=T 
        ti_list,xout,index,condM=self.Taylor_integration_timeList(ti_list,x0,ftol=ftol,max_iterations=max_iterations)
        if shape(ti_list)[0]< shape(xout)[0]:
            return xout[:shape(ti_list)[0],:],ti_list,index[:shape(ti_list)[0]],condM[:shape(ti_list)[0]]
        else:
            return xout,ti_list,index,condM

    def Taylor_Step(self,x0,h):
        '''
        Taylor_Step computes an approximation for the Taylor series X0 at t_0+h if the Taylor series X0 at t_0 and h are given.
            :param: 
                * **x0(t)** - ndarray with Taylor coefficents
                * **h** - step size
            :return:
                * **y** - x0(t+h)
        '''
        #print 'Taylor_Step: ================================'
        dim=int(shape(x0)[0]/(self.K+1))
        X0=reshape(x0,(self.K+1,dim)).copy()
        
        #print 'Taylor_Step:  x0,  X0 ',x0,'\n\n',X0
        factor=npzeros(self.K)
        shiftindex=int(self.index)#-1 # ohne -1??????
        if shiftindex == -1:
            if self.info_output[2]>'0':
                print ('Taylor_Step: Step after singularity, try to continue. Check the point t!')
            shiftindex=0
        for j in range(0,self.K):
            factor[j]=1.
        for j in range(0,self.K-shiftindex):        
            for i in range(j+1,self.K+1-shiftindex):
                X0[j,:]= X0[j,:]+X0[i,:]*h**(i-j)*factor[i-j-1]
                factor[i-j-1]=factor[i-j-1]*(i+1.)/(j+1.)
                
        y= reshape(X0,(self.K+1)*dim)
        #exit()
        return y[:-dim]
           
        
    def consistentStep(self,TaylorOrder,alpha,x0,ti,augmentation=False,ftol=1e-8,disp=True,maxiter=100,info_output='001'):
        """
        consistentStep computes consistent values of the DAE example in the timepoint ti such that || P(x(ti)-alpha) ||_2 -> min! using the modul minimize with the method SLSQP
            :param: 
                * **self**  - Taylor Object of the type TaylorIntegration
                * **TaylorOrder** - integer, order of the highest derivative of Taylor coefficients
                * **alpha** - reference vector as ndarray of size dim = dimension of the DAE
                * **x0** - ndarray containing initial guess of length dim*TaylorOrder
                * **ti** - timepoint
                * **augmentation** 
                    * if False: using the TaylorOrder
                    * if True:  augmentation of number of Taylor coefficients starting with 1 up to TaylorOrder+1
                * **ftol** = 1e-8 - (parameter for python module *minimize*)
                * **disp** = True - (parameter for python module *minimize*)
                * **maxiter** = 100 - (parameter for python module *minimize*): maximal number of iterations
                * **info_output** - additional component 'c', string 'abc' controls the information output, 0 no output up to 3 most information
                    a - level for linear algebra, like SVD, rank determination etc.\n
                    b - level for calculation of consistent initial values\n
                    c - level for integration\n
                * **info_level** = 0 - integer, higher level implies more information, 0 ... 3
                                       If info_level = 0 no output with the exception of error messages.\n
                                       If info_level > 0 then two log-files are generated: *example_name*\ _array_\ *date*\ .log and *example_name*\ _consi_\ *date*\ .log \n
                                       *array* contains Jacobian matrices, projectors, ranks, SVD values etc.\n
                                       *consi* contains values related to the computation of consistent initial values and the solution in full accuracy (16 digits)
            :raised:
                * **exceptions**: 
                    * MinimizationNotSuccessful
                    * ValueError
            :return: 
                * **solution** as ndarray of length dim*(TaylorOrder+1) containing values for x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(TaylorOrder)`\ (t0)/(TaylorOrder)!
                    **Caution**: Only  x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(s)`\ (t0)/s!  are consistent for s = TaylorOrder+1-index (see [1])
                * **TaylorEx** Object of class TaylorDAE of the example (to be used in index_Condition) 
        """
        
        #print 'consistentStep type(self)', type(self)
        info_level=max(info_output)
        example=self.example
        #print 'example in consistentStep',type(example)
        K = TaylorOrder+1
        dimension=alpha.shape
        dim=dimension[0]
        # check of the right length of x0
        if x0.size != TaylorOrder*dim:
            print ('consistentStep: Wrong length of x0, should be TaylorOrder*alpha.size = ',TaylorOrder*dim,' but it is',x0.size,'.\n')
            raise ValueError('Wrong length of x0')
        if augmentation:
            X=x0[:dim]
            
            for K_temp in range(2,K+1):
                K_tempmdim=K_temp*dim
                if K_temp==K:
                    X=append(X,npzeros(dim))
                else:
                    X=append(X,x0[K_tempmdim-dim:K_tempmdim])
                #X[:dim]=alpha
                if info_output[1]>'0':
                    print ('K_temp= ',K_temp,'\n\nX = \n\n',X.reshape(K_temp,dim))
                if hasattr(example,'svdtol'):
                    tolexample=example.svdtol
                else:
                    tolexample=None    
                TaylorEx = TaylorDAE(example, K_temp,tol=tolexample,info_output=info_output)
                if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                
                    if info_level>'0':
                        print ('consistentStep: A function u with additional conditions is used.')
                    logging.getLogger("consi").log(20,'\nconsistentValues: A function u with additional conditions is used.')
                
                CoeffMatrix=JacobianDerivativeArray(X,TaylorEx,ti,K_temp,dim)
                G0=CoeffMatrix[:dim,dim:2*dim]
                Gm=pinv(G0)
                P0=dot(Gm,G0)
                #print 'example in augmentation',type(example)
                logging.getLogger("array").log(20,'\nconsistentStep: P0 =\n\n'+str(P0))
                res = minimize(self.func_int, X, args=(P0,alpha,self), jac=self.func_int_deriv, constraints=constraint(TaylorEx,ti,K_temp,dim),\
                                method='SLSQP', options={'disp': disp,'maxiter':maxiter,'ftol':ftol})
                if res.success and TaylorEx.normDerivativeArray > 10*ftol:
                    raise MinimizationNotSuccessful()
                #print 'res',res
                if disp:
                    print ('minimize success =',res.success,' status =',res.status)
                if res.success:
                    X=res.x
                else:
                    raise MinimizationNotSuccessful()
        else:
            X=append(x0,npzeros(dim))
            if hasattr(example,'svdtol'):
                tolexample=example.svdtol
            else:
                tolexample=None
            TaylorEx = TaylorDAE(example, K,tol=tolexample,info_output=info_output)
            if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                    if info_output[1]>'0':
                        print ('consistentValues: A function u with additional conditions is used.')
                    logging.getLogger("consi").log(30,'\nconsistentValues: A function u with additional conditions is used.')
            CoeffMatrix=JacobianDerivativeArray(X,TaylorEx,ti,K,dim)
            G0=CoeffMatrix[:dim,dim:2*dim]
            Gm=pinv(G0)
            P0=dot(Gm,G0)
            logging.getLogger("consi").log(10,'\nConsistentValues: info_output[1] = '+info_output[1])
            logging.getLogger("array").log(20,'\nConsistentValues: P0 =\n\n'+str(P0))
            
            res = minimize(self.func_int, X, args=(P0,alpha,self), jac=self.func_int_deriv, constraints=constraint(TaylorEx,ti,K,dim),\
                           method='SLSQP', options={'disp': disp,'maxiter':maxiter,'ftol':ftol})
            #print 'res',res
            if res.success and TaylorEx.normDerivativeArray > 10*ftol:
                print ('consistentValues: Norm of the derivative array to big = {:9.6e}'.format(TaylorEx.normDerivativeArray))
                raise MinimizationNotSuccessful()
            if disp:
                print ('minimize success =',res.success,' status =',res.status)
            if not res.success:
                raise MinimizationNotSuccessful()
        
        logging.getLogger("consi").log(30,'\nConsistentValues: '+res.message)
        resString=''
        for step in range(K):
            set_printoptions(precision=16)
            resString=resString+array_str(res.x[step*dim:step*dim+dim], max_line_width=dim*28)+'\n'
        equals=    '\n===============================================================\n'
        logging.getLogger("consi").log(30,'\n'+equals+'ConsistentValues: Computed consistent values at t = '+str(ti)+equals+'\n'+resString)
        return res.x,TaylorEx
              
        
    def consistentStepIndex(self,TaylorOrder,alpha,x0,t0,augmentation=False,ftol=1e-8,disp=False,maxiter=100,info_output='001'):
        """
        consistentStepIndex wraps the moduls consistentValues and index_Condition.
            :param: 
                * **self**  - Taylor Object of the type TaylorIntegration
                * **TaylorOrder** - integer, order of the highest derivative of Taylor coefficients
                * **alpha** - reference vector as ndarray of size dim = dimension of the DAE
                * **x0** - ndarray containing initial guess of length dim*TaylorOrder
                * **t0** - timepoint
                * **augmentation** = False
                    * if False: using the TaylorOrder
                    * if True:  augmentation of number of Taylor coefficients starting with 1 up to TaylorOrder+1
                * **ftol** = 1e-8 - (parameter for python module *minimize*)
                * **disp** = False - (parameter for python module *minimize*)
                             If *disp* = True also results of the iteration of the minimizer are displayed.
                * **maxiter** = 100 - (parameter for python module *minimize*): maximal number of iterations
                * **info_output** - additional component 'c', string 'abc' controls the information output, 0 no output up to 3 most information
                    a - level for linear algebra, like SVD, rank determination etc.\n
                    b - level for calculation of consistent initial values\n
                    c - level for integration\n 
            :return: 
                * **solution** as ndarray of length dim*(TaylorOrder+1) containing values for x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(TaylorOrder)`\ (t0)/(TaylorOrder)!
                    **Caution**: Only  x(t0), x'(t0),x"(t0)/2!, ..., x :sup:`(s)`\ (t0)/s!  are consistent for s = TaylorOrder+1-index (see [1])
                * **index** of the DAE in the computed consistent point
                * **cond** - condition number of the index computation
                * **computation_robust** - If True -> The relation TaylorOrder+1-index-s==0 is fulfilled
        """
       # print 'consistent Integration TaylorObject',type(TaylorObject)
        info_level=max(info_output)
        dimension=alpha.shape
        n=dimension[0]
        stars='*'*57+'\n'
        try:
            if info_output[1]>'1':
                print ('\n',stars,'consistentStepIndex: Computation of consistent solution\n',stars)
            solution,TaylorEx=self.consistentStep(TaylorOrder,alpha,x0,t0,augmentation=augmentation,ftol=ftol,disp=disp,maxiter=maxiter,\
                                                                                                                                       info_output=info_output)
            if info_output[1]>'0':
                print ('\n\n',stars,'consistentStepIndex: Consistent solution at t = ',t0,'\n',stars,'\n',solution.reshape(TaylorOrder+1,n))
            index,cond,s=index_Condition(TaylorEx,n,solution,t0)
            if s==-1:
                logging.getLogger("consi").log(30,'\nconsistentValuesIndex: DAE may be singular!\n'\
                                             +'                       The computed minimization solution does not represent consistent values,\n'\
                                             +'                       index and condition number could not be computed.')
                return solution,NaN,cond,False
            if index==0:
                s+=1 # see DAE_TaylorDAE.Check_s_full
            
            computation_robust= TaylorOrder+1-index-s==0   
            if info_output[1]>'0':    
                print ('\nconsistentStepIndex: index = ',index,' D = ',(TaylorOrder+1),'  s = ',s, ' (D - index  == s ?) = ',computation_robust,\
                      '\n                       condition number of the derivative array considered in the final step of index determination = {0:10.2e}'.format(cond))                                
            logging.getLogger("consi").log(20,'\nconsistentValuesIndex: \n'+'Number of Taylor coefficents = '\
                                         +str(TaylorOrder+1)+'  index mu = '+str(index)+'  cond = {0:10.2e}'.format(cond)\
                                         +'  Coefficient matrix is '+str(s)+'-full')
            if TaylorOrder+1-index != s:
                if info_output[1]>'0':
                    messagepart1='consistentStepIndex: The relation D-mu == s is not fulfilled. Check the condition number and the validity of the index determination!!!!!\n'
                    messagepart2=' or in the array-file and maybe adapt the parameter svdtol in the example.'
                    if info_level<'3':
                        message=messagepart1+\
                                '                       Set info_output[0]="3" and check the singular value S(r) in the terminal output'+messagepart2
                                
                    else:
                        message=messagepart1+\
                                '                       Check the singular value S(r) in the terminal output'+messagepart2
                    print (message)
                    
                    logging.getLogger("consi").log(30, message)
            return solution,index,cond,computation_robust
        
        except MinimizationNotSuccessful as ex:
            print (ex,'\nconsistentStepIndex: InitDAE cannot compute a consistent solution!')
            if 'TaylorEx' in locals(): # this is the case when consistentValues is aborted by Exception
                if hasattr(TaylorEx.example,'function_u') and TaylorEx.example.function_u:
                    print ('consistentStepIndex: ----- Additional function u is used, it maybe that the condition(s) in function u are not performable.')
            print ('consistentStepIndex  ex.index',self.index)      
            return None,NaN,NaN,None
        except ValueError as ex:
            print (ex,'\n\nconsistentStepIndex: InitDAE cannot compute a consistent solution!')
            return None,NaN,NaN,None
        except LittleNumberOfTaylorCoefficients as ex:
            logging.getLogger("consi").log(30,'\nconsistentStepIndex: Number of Taylor coefficients too small!\n'\
                                             +'                       The computed minimization solution does not represent consistent values,\n'\
                                             +'                       index and condition number could not be computed.')
            print ('consistentStepIndex: ',ex)
            raise
        except SingularDAE as ex:
            print ('consistentStepIndex: ',ex)
            if solution is None:
                return None,NaN,NaN,False
            else:
                return solution,NaN,NaN,False
            
            
    @staticmethod        
    def func_int(X,P,alpha,Object):
        """
        Objective function  ||P(x_0-alpha)||_2, where X = (x_0,x_1,...)
            :param: 
                * **X** - ndarray of Taylor variable
                * **P** - orthogonal projector
                * **alpha** - initial guess
                * **Object** - Taylor object
            :return:
                * **sqrt(yty)** - ||y||_2 with y = P(x0-alpha)
        """
        n=size(P,0)
        x0=X[:n]
        method=Object.method        
        if method > 0:
            x0=Object.Taylor_Method_x0i(X,n)
        
        y=dot(P,x0-alpha)
        yty=dot(y,y)
        
        #print 'func_int, type',type(ex)
        logging.getLogger("consi").log(20,'\t\t\t\t func: sqrt(yty) {0:10.2e}'.format(sqrt(yty)))
        if Object.info_output[1]>'1':
            print ('\tfunc: sqrt(yty) {0:10.2e}'.format(sqrt(yty)))#,y    
        return sqrt(yty)
    
    @staticmethod
    def func_int_deriv(X,P,alpha,Object):
        """ 
        Derivative of objective function 
            :param: 
                * **X** - ndarray of Taylor variable
                * **P** - orthogonal projector
                * **alpha** - 'initial value'
                * **Object** - Taylor object
            :return:
                * **deriv** - 
        """
        n=size(P,0)
        #print 'dimension ',n
        deriv=npzeros(X.size)
        x0=X[:n]
        #-------------------------------------------------------------- 
        method=Object.method        
        if method > 0:
            x0=Object.Taylor_Method_x0i(X,n)
        #--------------------------------------------------------------
        y=dot(P,x0-alpha)
        
        yty=dot(y,y)
#        print 'func_int_deriv: yty',yty,y
        
        threshold=2.3e-16
        
        if sqrt(yty)>threshold:
            aux=y/sqrt(yty)
        else:
            yaux=dot(P,ones(size(y)))
            ytyaux=dot(yaux,yaux)
            aux=yaux/sqrt(ytyaux) #Unity vector!
        
        deriv[:n]=aux
        
        if method==1:
            for k in range(X.size/n):

                deriv[k*n:(k+1)*n]=aux*(-Object.h)**k
        elif method==2:
            for k in range(X.size/n):

                deriv[k*n:(k+1)*n]=aux*(-Object.h/2)**k
        elif method>2:
             for k in range(Object.ki+1):
                deriv[k*n:(k+1)*n]=aux*((-Object.h)**k*Object.HOPweightsI[k])
        
        return deriv

    def Taylor_Method_alpha(self,Xj,n,h):
        """ 
        Taylor_Method_alpha computes alpha according to the chosen integration method.
            :param:
                * **X** - ndarray of Taylor variable
                * **n** - dimension of DAE
                * **h** - current stepsize
            :return:
                * **alpha** -  
        """
        method=self.method  
        if method == 0: #Explicit Taylor
            aux=self.Taylor_Step(Xj,h)
            alpha=aux[:n]
        elif method == 1: #Fully implicit Taylor        
            alpha=Xj[:n]
        elif method == 2: # Two halfstep
            xH=self.Taylor_Step(Xj,h/2)
            alpha=xH[:n]
        else: #  HOP
            alpha=self.HOP_Evaluation(Xj,'E') 
        
        return alpha
    
    #@staticmethod
    def Taylor_Method_x0i(self,X,n):
        """ 
        Compute x0i according to the chosen Taylor Method
            :param:
                * **X** - ndarray of Taylor variable
                * **n** - dimension of DAE
            :return:
                * **x0i**
        """
        method=self.method
        
        if method == 1: #Fully implicit Taylor        
            xH=self.Taylor_Step(X,-self.h)
            x0i=xH[:n]
            
        elif method == 2: # Two halfstep

            xH=self.Taylor_Step(X,-self.h/2)
            x0i=xH[:n]
            
        else: #  HOP
           # print 'HOP'
            x0i=self.HOP_Evaluation(X,'I')
        
        return x0i
        
    #@staticmethod    
    def HOP_Evaluation(self,x0,E_I):
        ''' 
        Computes xx according to HOP-formulas (e/i)
            :param:
                * **x0** - ndarray
                * **h** - step size
                * **E_I** - decides whether we compute x_e or x_i
                    for E_I=='E': x_e\n
                             'I': x_i\n         
                * **xx** (for x_e or x_i)
        '''

        #print 'HOP_Evaluation: ================================'
        dim=int(shape(x0)[0]/(self.K+1))
        #print 'HOP_Evaluation: dim : ',shape(x0),self.K,dim
        X0=reshape(x0,(self.K+1,dim)).copy()
        #print 'HOP_Evaluation:  x0,  X0 ',x0,'\n\n',X0
        
        if E_I=='E':
            kk=self.ke
            ww=self.HOPweightsE
            hh=self.h
        elif E_I=='I':
            kk=self.ki
            ww=self.HOPweightsI
            hh=-self.h
        else:
            print ("HOP_Evaluation: error - second argument must be 'E' or 'I'")
            exit()
        # Horner scheme
        xxx=X0[kk,:]*ww[kk]
        for j in range(kk-1,-1,-1):
            xxx=xxx*hh+X0[j,:]*ww[j] 
        
        return xxx
    
    def plot_figure(self,tlist,x,mylabel,xlab='t',ylog=False):
        """
        plot_figure supports the output of graphics of results stored in ndarrays of the form [ number of timepoints, dim of the DAE]
            :param:
                * *tlist* - list of timepoints
                * *x* - list of solution at the timepoints
                * *mylabel* - string for labeling x
                * *xlab* - label of x-axis
                * *ylog* - switch for logarithmic abscissa
            :return: no output, plot opens
        """
        #from matplotlib import use,get_backend
        #print 'backend ',get_backend()
        #use('WxAgg')
        #use('Gtk3Agg')
        #if not str(get_backend())=='TkAgg':
        #    use('TkAgg')
        #use('Qt4Agg')
        from matplotlib.pyplot import plot,show,legend,figure,xlabel,yscale,ylabel,title,yticks
        dim=x.ndim
        if dim==1:
            n=1
        else:
            n=x.shape[1]
        figure(0)
        if dim==1 or x.shape[1]==1:
            #print 'dim=1'
            #print 't,x ',tlist,x
            plot(tlist,x,label=mylabel)
        else:
            for i in range(n):
                label=mylabel+str(i)
                plot(tlist,x[:,i],label=label)
        
        legend(loc='upper left', shadow=True)
        title(self.example.example_name)
       
        if ylog:
            yscale('symlog')
            #yticks([tlist[0],tlist[-1]])
        xlabel(xlab)    
        show()

# we are outside of the class now    
def compute_finite_differences(solution_list):
    """
    compute_finite_differences delivers a list of t_i, x(t_i), x'(t_i).\n
    The first derivative is approximated by finite differences of second order
        :param: 
            * **solution_list** - array with columns of time point, x(1),...x(M-1)
        :return:
            * **derivative_list** -  ndarray with x(t_i), x'(t_i), used in adapt_solution_list
    """
    #print "in compute_solution"
    D=2
    Z=shape(solution_list)
    #print "shape(solution_list) = ",Z
    # number of timepoints
    P=Z[0]
    # DAE dimension+1
    M=Z[1]
    Mm1=M-1
    
    derivative_list=npzeros((D,P,Mm1))
    
    Pm1=P-1
    for i in range(1,Pm1):
        derivative_list[0,i,:]=solution_list[i,1:]
        #print 'solution list shape =',shape(solution_list[i,1:]),type(solution_list[i,1:])
        derivative_list[1,i,:]=(solution_list[i+1,1:]-solution_list[i-1,1:])/(solution_list[i+1,0]-solution_list[i-1,0])
        
    derivative_list[0,0,:]=solution_list[0,1:]    
    derivative_list[1,0,:]=(-3*solution_list[0,1:]+4*solution_list[1,1:]-solution_list[2,1:])\
                    /(solution_list[Pm1,0]-solution_list[Pm1-2,0])    
    derivative_list[0,Pm1,:]=solution_list[Pm1,1:]
    #derivative_list[1,0,Pm1,:]=(-3*solution_list[Pm1,1:]+4*solution_list[Pm1-1,1:]-solution_list[Pm1-2,1:])\
    #               /(solution_list[Pm1,0]-solution_list[Pm1-2,0]) 
    derivative_list[1,Pm1,:]=(3*solution_list[Pm1,1:]-4*solution_list[Pm1-1,1:]+solution_list[Pm1-2,1:])\
                    /(solution_list[Pm1,0]-solution_list[Pm1-2,0])                 
    #print "out compute_solution",derivative_list
    return derivative_list

# we are outside of the class now
def adapt_solution_list(path,K):
    ''' 
    adapt_solution_list computes a list of Taylor function values up to order K, where the values of the\n
    solution in the file "path" are used and the first derivative approximated by finite differences.\n
    The other Taylor coefficients from 2 to K are set to zero.
        :param:
            * **path** - (relative) path to the file with solution in the form t, x(t) 
            * **K** - number of Taylor coefficients
        :return:
            * **Taylor object** with solution, used in check_solution_along_path
    '''
    liste = loadtxt(path)
    dim=shape(liste)
    #print "Dimension of data file= "+str(dim), liste
        
    # number of time points
    time=dim[0]
       
    # approximation of the first derivative by finite differences
    derivative_list=compute_finite_differences(liste) 
    #print 'adapt_solution_list  ',derivative_list
    n=dim[1]-1#shape(derivative_list)[1]
    
    ti=float(liste[0,0])
    Time = MyUTPM.taylor_init(K, ti)
    
    zzt= MyUTPM(zeros((K, 1, n, time)))
    for i in range(time): # main loop 
        ti=float(liste[i,0])
        Time.data[0,0]=ti
        #print 'zzt ',shape(zzt.data),zzt.data[0,0,:,i],'\n\n derivative_list  ',shape(derivative_list),derivative_list[0,i,:]
        zzt.data[0,0,:,i]=derivative_list[0,i,:]
        zzt.data[1,0,:,i]=derivative_list[1,i,:]
        
    return zzt
 

#######################################
   
if __name__ == "__main__":

    from algopy import sin,cos, exp
    from matplotlib.pyplot import plot,show
    from AD_examples.example_ind2_sing import Example_ind2_sing
    from numpy import pi

    ex=Example_ind2_sing()
    K=14
    #K=5
    Taylorex=TaylorIntegration(ex,K,tol=None,info_output='001',method=4)
    
    def x(t):
        xt = zeros(2, dtype=t)
        xt[0]=sin(t)
        xt[1]=1+t**2
        return xt
    a=.0
    b=2.0
    time_points=100
    condition,index_list=Taylorex.check_solutionpath(x,a,b,time_points=time_points)
    print ('check_solutionpath++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    #print 'condition index ',condition, index_list    
    plot(condition[:,0],condition[:,1])
    show()
    
    print ('check_solution_along_path++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    path=ex.path
    print ('path = ',path)
    tliste,index,condM=Taylorex.check_solution_along_path(path)

    plot(tliste,condM)
    show()
    #exit()
    t0=0.0#pi/4
    alpha=array([0.,1.])
    DAE_index=-1
    h=0.1
    NumSteps=10
    
    T=2.
    # Test:
    print ('Taylor_integration_constant_h++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    NumSteps=20
    #x0=alpha
    yxout,tlist,indexList,condList=Taylorex.Taylor_integration_constant_h(t0,alpha,T,NumSteps,ftol=1e-6,max_iterations=20)
        
    
#     if Taylorex.method>2:    
#         print 'HOPweightE ',Taylorex.HOPweightsE
#         print 'HOPweightI ',Taylorex.HOPweightsI    
    
    #print indexList
    #print condList
    Taylorex.plot_figure(tlist,yxout,'y',xlab='t',ylog=False)
    Taylorex.plot_figure(tlist,condList,'condition',xlab='t',ylog=False)
    
    print ('Taylor_integration_adaptiveStep++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
    print ('alpha= ',alpha)
    yxout,tlist,indexList,condList=Taylorex.Taylor_integration_adaptiveStep(t0,T,alpha,disp=False,tol=1e-14,max_iterations=100)
    if yxout is None:
        exit()
    Taylorex.plot_figure(tlist,yxout,'y',xlab='t',ylog=False)
    Taylorex.plot_figure(tlist,condList,'condition',xlab='t',ylog=False)

    
 